;;; Some ATD local extensions for epoch.  K. C. Laprade - 27 Aug 90.

(provide 'epoch-local)

;;; Use a nice arrow cursor.
(cursor-glyph 68)
(cursor-glyph 68 (minibuf-screen))
(push '(cursor-glyph . 68) epoch::screen-properties)

;;; Compatibility with X GNUemacs:
(fset 'x-set-bell 'epoch::set-bell)
(fset 'x-get-cut-buffer 'epoch::get-cut-buffer)
(fset 'x-store-cut-buffer 'epoch::store-cut-buffer)

;;; For ebuff-menu:
(defvar electric-buffer-menu-mode-mouse-map (create-mouse-map))
(define-mouse electric-buffer-menu-mode-mouse-map mouse-left mouse-down 'mouse::set-point)
(define-mouse electric-buffer-menu-mode-mouse-map mouse-left mouse-up
  (function (lambda (arg) (mouse::set-point arg) (Electric-buffer-menu-select))))
(define-mouse electric-buffer-menu-mode-mouse-map mouse-middle mouse-down 'mouse::set-point)
(define-mouse electric-buffer-menu-mode-mouse-map mouse-middle mouse-up
  (function (lambda (arg) (mouse::set-point arg) (Electric-buffer-menu-select))))
(define-mouse electric-buffer-menu-mode-mouse-map mouse-right mouse-down 'mouse::set-point)
(define-mouse electric-buffer-menu-mode-mouse-map mouse-right mouse-up
  (function (lambda (arg) (mouse::set-point arg) (Electric-buffer-menu-select))))
(defvar electric-buffer-menu-mode-hook nil)
(if (not (listp electric-buffer-menu-mode-hook))
    (setq electric-buffer-menu-mode-hook (list electric-buffer-menu-mode-hook)))
(setq electric-buffer-menu-mode-hook (cons 'epoch-electric-buffer-menu-mode-hook electric-buffer-menu-mode-hook))
(defun epoch-electric-buffer-menu-mode-hook ()
  (use-local-mouse-map electric-buffer-menu-mode-mouse-map))

;;; Autoloads:
(autoload 'mouse-helper "mouse-help" "Popup window with mouse bindings (not loaded)." t)
(autoload 'describe-mouse "mouse-help" "Describe mouse button (not loaded)." t)
(autoload 'describe-mouse-briefly "mouse-help" "Describe mouse button briefly (not loaded)." t)

;;; More functions:

(defun one-screen-p () "Returns non-nil it there is exactly one screen."
  (if (cdr (screen-list)) nil t))
  
;; Some additional X functions to bind to keys (PUT and GET).
(defun x-copy-region (beg end &optional kill)
  "Copy text between point and mark into window system cut buffer.
Save in Emacs kill ring also.  With optional arg, delete it as well."
  (interactive "rP")
  (if kill (barf-if-buffer-read-only))
  (x-store-cut-buffer (buffer-substring beg end))
  (copy-region-as-kill beg end)
  (if kill (delete-region beg end)))

(defun x-cut-region (beg end)
  "Copy text between point and mark into window system cut buffer.
The region is killed and saved in the Emacs kill ring."
  (interactive "r")
  (x-store-cut-buffer (buffer-substring beg end))
  (copy-region-as-kill beg end)
  (delete-region beg end))

(defun x-paste-point ()
  "Insert window system cut buffer contents at point."
  (interactive)
  (setq this-command 'yank)
  (push-mark (point))
  (insert (x-get-cut-buffer)))

;;; Support for generic function key binding (by Ken Laprade):
;;;
(defvar function-key-prefix-string "\C-X*"
  "Prefix for function key definitions created by define-function-key.")

(defun define-function-key (key def &optional epoch-def)
  "Globally define a generic function key for epoch.  KEY is a string naming
the key (as known to X) with optional `M', `S', or `C' prefixes specifying
meta, shift and control qualifiers.  Prefixes are separated from the key
name by a hyphen.  DEF is the key definition.  It may be any valid
ddefinition for define-key.  If it is a string, it will be directly bound
to the key using rebind-key.  If the optional EPOCH-DEF is supplied, it
is used rather than def."
  (interactive "sFunction key name: \nxDefinition: ")
  (if epoch-def (setq def epoch-def))
  (let* ((i (string-match "-" key))
	 (keyname (let ((name (substring key (if i (1+ i) 0))))
		    (cond ((string-match "R8" name) "Up")
			  ((string-match "R10" name) "Left")
			  ((string-match "R12" name) "Right")
			  ((string-match "R14" name) "Down")
			  (t name))))
	 (meta (string-match "M.*-" key))
	 (shift (string-match "S.*-" key))
	 (control (string-match "C.*-" key))
	 quals)
    (if meta (setq quals (append '(meta) quals)))
    (if control (setq quals (append '(control) quals)))
    (if shift (setq quals (append '(shift) quals)))
    (or quals (setq quals 0))
    (if (stringp def)
	(rebind-key keyname quals def)
      ;; Make an intermediate binding for the definition.
      (let* ((keytype (cond ((string-match "L\\([1-9]\\|10\\)$" key) "L")
			    ((string-match "F\\([1-9]\\)$" key) "F")
			    ((string-match "R\\([1-9]\\|1[0-5]\\)$" key) "R")
			    (t keyname)))
	     (keynum (and (match-beginning 1)
			  (string-to-int (substring key (match-beginning 1) (match-end 1)))))
	     (binding (concat function-key-prefix-string
			      (if meta "M") (if shift "S") (if control "C")
			      keytype
			      (if keynum (format "%02d" keynum)))))
	(rebind-key keyname quals binding)
	(global-set-key binding def)))))

;;; Make the title and icon name change when the buffer is set.
;;; This uses an undocumented feature that SMK claims crashes Epoch.
;;; However, AMC says it is fixed and works just fine.

(defvar inhibit-rename-screen nil "List of screens not to rename on select")
(defvar title-add-sys-name t "Add system name in screen title if true.")
(defvar icon-add-sys-name nil "Add system name to icon name if true.")

(defun amc-set-name (win)
  (let* ((buff (window-buffer win))
	 (name (or (and buff (buffer-name buff)) "# No-Name #"))
	 (name-with-sys (concat name (sys-name)))
	 (scr (epoch::screen-of-window win)))
    (if (and
	 buff
	 (buffer-file-name buff)
	 (not (memq scr inhibit-rename-screen)))
	(progn
	  (epoch::title (if title-add-sys-name name-with-sys name) scr)
	  (epoch::icon-name (if icon-add-sys-name name-with-sys name) scr)))))

(setq set-window-buffer-hook 'amc-set-name)

(defun add-inhibit-rename-screen (screen)
  "Add SCREEN to the inhibit-rename-screen list.  Also clean out any
non-existing screens from the list."
  (or (memq screen inhibit-rename-screen)
	  (setq inhibit-rename-screen (cons screen inhibit-rename-screen)))
  (let ((l inhibit-rename-screen))
    (while (cdr l)
      (if (and (screen-p (car (cdr l)))
	       (get-screen-id (car (cdr l))))
	  (setq l (cdr l))
	(setcdr l (cdr (cdr l)))))))

(defun screen-attributes (geometry name &rest others)
  "Return an alist of screen attributes suitable for create-screen.
Includes GEOMETRY and title and icon name based on TITLE and all other
arguments which should be cons cells."
  (append (list (cons 'geometry geometry)
		(cons 'title (concat name (if title-add-sys-name (sys-name))))
		(cons 'icon-name (concat name (if icon-add-sys-name (sys-name)))))
	  others))


;;; For calculator:

(defvar calc-geometry "78x9+447+70"
  "*Geometry for calculator window created by popup-calc-screen.")
(defvar gnucalc-screen nil)

(defun popup-calc-screen ()
  "Startup a calculator on a new screen.  Iconify calculator if it is the
current screen."
  (interactive)
  (if (equal (current-screen) gnucalc-screen)
      (if (and (epoch::get-buffer-window "*Calculator*")
	       (equal (screen-of-window (epoch::get-buffer-window "*Calculator*")) (current-screen)))
	  (iconify-screen)
	(delete-other-windows)
	(full-calc))
    (if (and (screen-p gnucalc-screen) (get-screen-id gnucalc-screen)) t
      (setq gnucalc-screen (create-screen nil (screen-attributes calc-geometry "GNUcalc")))
      (add-inhibit-rename-screen gnucalc-screen))
    (cursor-to-screen (mapraised-screen (select-screen gnucalc-screen)))
    (or (eq (buffer-major-mode) 'calc-mode)
	(full-calc))))

(defun mouse-calc-grab-thing (arg)
  "Yank the region highlighted by grab-button into the calculator.
Intended as an up button following mouse-select-thing as a down button."
  (if (and (> mouse::clicks 0)		; Abort if mouse moved.
	   (button-buffer grab-button))
      (save-excursion
	(set-buffer (button-buffer grab-button))
	(let ((str (buffer-substring (button-start grab-button)
				     (button-end grab-button))))
	  (require 'calc "calc/calc")
	  (calc-select-buffer)
	  (calc-do
	   (function (lambda ()
		       (calc-pop-push-record-list
			0 "grab"
			(if (stringp str)
			    (let ((val (math-read-exprs (calc-clean-newlines str))))
			      (if (eq (car-safe val) 'error)
				  (error "Bad format in calc data")
				val))))))))))
  (if (button-buffer grab-button)
      (delete-button grab-button))
  (epoch::redisplay-screen))

;; Copied from calc-yank.el because it has no provide:
(defun calc-clean-newlines (s)
  (cond
      ;; Omit leading/trailing whitespace
   ((or (string-match "\\`[ \n\r]+\\([^\001]*\\)\\'" s)
	(string-match "\\`\\([^\001]*\\)[ \n\r]+\\'" s))
    (calc-clean-newlines (math-match-substring s 1)))
   ;; Convert newlines to commas
   ((string-match "\\`\\(.*\\)[\n\r]+\\([^\001]*\\)\\'" s)
    (calc-clean-newlines (concat (math-match-substring s 1) ","
				 (math-match-substring s 2))))
   (t s)))


;;; For mail:

(defvar mail-geometry "80x40+0+100"
  "*Geometry for mail window created by \\[mail-other-screen].")
(defvar mail-screen nil)

(defun mail-other-screen (&optional noerase to subject in-reply-to cc replybuffer)
  "Just like mail, but popup a new screen."
  (interactive "P")
  (require 'sendmail)
  (define-key mail-mode-map "\C-c\C-c" 'mail-send-and-exit-screen)
  (set-buffer (get-buffer-create "*mail*"))
  (setq default-directory (expand-file-name "~/"))
  (auto-save-mode auto-save-default)
  (mail-mode)
  (and (not noerase)
       (or (not (buffer-modified-p))
	   (y-or-n-p "Unsent message being composed; erase it? "))
       (progn (erase-buffer)
	      (mail-setup to subject in-reply-to cc replybuffer)
	      t))
  (if (and (screen-p mail-screen) (get-screen-id mail-screen)) t
    (setq mail-screen (create-screen "*mail*" (screen-attributes mail-geometry "Mail")))
    (add-inhibit-rename-screen mail-screen))
  (cursor-to-screen (mapraised-screen (select-screen mail-screen))))

(defun mail-send-and-exit-screen (arg)
  "Send message like mail-send, then, if no errors, exit from mail buffer.
Prefix arg means don't delete this screen (only deletes mail-screen)."
  (interactive "P")
  (mail-send)
  (bury-buffer (current-buffer))
  (if (and (not arg)
	   (cdr (screen-list))
	   (equal (current-screen) mail-screen))
      (delete-screen (unmap-screen (current-screen)))
    (if (and (not arg)
	     (not (one-window-p))
	     (save-excursion
	       (set-buffer (window-buffer (next-window (selected-window) 'not)))
	       (eq major-mode 'rmail-mode)))
	(delete-window)
      (switch-to-buffer (other-buffer (current-buffer))))))


;;; Other stuff:

(defun scroll-other-screen-up-line (arg) "Scroll the other screen up ARG lines."
  (interactive "p")
  (if (one-screen-p) (error "There is no other screen"))
  (select-screen (next-screen))
  (or (pos-visible-in-window-p (point-max))
      (scroll-up arg))
  (select-screen (prev-screen)))

(defun scroll-other-screen-down-line (arg) "Scroll the other screen down ARG lines."
  (interactive "p")
  (if (one-screen-p) (error "There is no other screen"))
  (select-screen (next-screen))
  (or (pos-visible-in-window-p (point-min))
      (scroll-down arg))
  (select-screen (prev-screen)))

(defun scroll-other-screen-up (arg) "Scroll the other screen up."
  (interactive "P")
  (if (one-screen-p) (error "There is no other screen"))
  (select-screen (next-screen))
  (or (pos-visible-in-window-p (point-max))
      (scroll-up arg))
  (select-screen (prev-screen)))

(defun scroll-other-screen-down (arg) "Scroll the other screen down."
  (interactive "P")
  (if (one-screen-p) (error "There is no other screen"))
  (select-screen (next-screen))
  (or (pos-visible-in-window-p (point-min))
      (scroll-down arg))
  (select-screen (prev-screen)))

(defun beginning-of-buffer-other-screen (arg)
  "Do beginning-of-buffer in the other screen."
  (interactive "P")
  (if (one-screen-p) (error "There is no other screen"))
  (select-screen (next-screen))
  (beginning-of-buffer arg)
  (select-screen (prev-screen))
  (message "Mark set in other screen"))

(defun end-of-buffer-other-screen (arg)
  "Do end-of-buffer in the other screen."
  (interactive "P")
  (if (one-screen-p) (error "There is no other screen"))
  (select-screen (next-screen))
  (end-of-buffer arg)
  (select-screen (prev-screen))
  (message "Mark set in other screen"))

(defun pop-point-other-screen (arg)
  "Do pop-point in the other screen."
  (interactive "P")
  (if (one-screen-p) (error "There is no other screen"))
  (select-screen (next-screen))
  (set-mark-command t)
  (select-screen (prev-screen))
  (message "Point popped in other screen"))

(defun pop-point () "Pop point."
  (interactive)
  (set-mark-command t)
  (if drag-button
      (progn (delete-button drag-button) (setq drag-button nil)))
  (message "Point popped"))


;;; Add a map event handler that gets rid of auto-unmap-screens when the
;;; last normal screen is closed and brings them back when any screen is
;;; opened.  auto-unmap-screens is a list of symbols that eval to screens.
;;; The last size and position is remembered for each window because
;;; mapping seems to resize screens to their creation size.

(defvar auto-unmap-screens nil
  "*List of symbols referring to screens that will be automatically unmapped
when the last normal screen is closed and remapped when any screen is opened.")
(defvar auto-unmap-screens-geometry nil
  "Alist of saved geometries for screens referenced by auto-unmap-screens.")

(defun normal-screens (screens)
  "Return a list of all screens in SCREENS that are not referenced by
auto-unmap-screens."
  (let (l)
    (while screens
      (let ((scrn (car screens))
	    isone)
	(mapcar
	 (function (lambda (scr)
		     (setq isone (or isone (eq scrn (eval scr))))))
	 auto-unmap-screens)
	(or isone
	    (setq l (append l (list scrn)))))
      (setq screens (cdr screens)))
    l))

(defun auto-map-event-handler (type value screen)
  "Map event handler.  Will unmap auto-unmap-screens if this is the last
normal screen being unmapped.  Will remap auto-unmap-screens if mapping a
screen.  Calls on-event::handler first."
  (on-event::handler type value screen)
  (if value
      ;; Mapping.
      (mapcar
       (function
	(lambda (scr)
	  (setq scr (eval scr))
	  (or (not (screen-p scr))
	      (eq scr screen)
	      (screen-mapped-p scr)
	      (progn
		(map-screen scr)
		;; Unmapped screens can forget their geometry, so we have to restore it.
		(let* ((geom (cdr (assq scr auto-unmap-screens-geometry))))
		  (if geom
		      (progn
			(change-screen-size (car geom) (nth 1 geom) scr)
			(move-screen (nth 2 geom) (nth 3 geom) scr))))))))
       auto-unmap-screens)
    ;; Unmapping.
    (or (normal-screens (screen-list))
	;; No more normal screens are mapped: unmap all others.
	(mapcar
	 (function
	  (lambda (scr)
	    (setq scr (eval scr))
	    (or (not (screen-p scr))
		(not (screen-mapped-p scr))
		;; Unmapped screens can forget their geometry, so we have to remember it.
		(let* ((entry (assq scr auto-unmap-screens-geometry))
		       (info (epoch::screen-information scr))
		       (border (nth 5 info))
		       (x (- (car info) border))
		       (y (- (nth 1 info) border))
		       (geom (list (epoch::screen-width scr)
				   (epoch::screen-height scr)
				   x y)))
		  (if entry
		      (setcdr entry geom)
		    (setq auto-unmap-screens-geometry
			  (append auto-unmap-screens-geometry
				  (list (cons scr geom)))))
		  (unmap-screen scr)))))
	 auto-unmap-screens))))
	
(setq auto-unmap-screens (append auto-unmap-screens (list '(minibuf-screen))))
(pop-event 'map)
(push-event 'map 'auto-map-event-handler)

(defun delete-screen-or-kill-emacs (arg)
  "Delete the current screen.  If this is the only screen, then
save buffers and kill emacs."
  (interactive "P")
  (if (delq (current-screen) (normal-screens (screen-list t)))
      (progn (delete-screen (current-screen))
	     (auto-map-event-handler nil nil nil))
    (if (y-or-n-p "Really exit? ")
	(save-buffers-kill-emacs arg)
      (message ""))))

;;; Disable suspending.
;;;
(put 'suspend-emacs 'disabled
     "Suspending a program running in an X window is silly
and you would not be able to start it again.  Just switch windows instead.\n")
(setq suspend-hook '(lambda () (error "Suspending an emacs running under X makes no sense")))
(substitute-key-definition 'suspend-emacs nil global-map)
(substitute-key-definition 'suspend-emacs nil esc-map)
(substitute-key-definition 'suspend-emacs nil ctl-x-map)
