;;; Personal setup for using epoch. - Ken Laprade - 15 Nov 90

(set-bell t)
(setq auto-raise-screen nil)
;; Raise minibuffer whenever a window is selected:
(push (function (lambda () (raise-screen (minibuf-screen)))) *select-screen-hook*)

(setq epoch-mode-alist
  (nconc (list
	  (cons 'Info-mode (list (cons 'geometry "80x40+0+257")))
	  (cons 'dired-mode (list (cons 'geometry "80x40+0+70")))
	  (cons 'text-mode (list (cons 'geometry "80x50+0+70"))))
	 epoch-mode-alist))

;;; To set geometry for screen generated by server.el:
(setq target-buffer (get-buffer-create " *target*"))
(save-excursion (set-buffer target-buffer)
		(text-mode))

;;; Changed/additional key sequences:
(global-set-key "\C-za" (definteractive (let ((name (concat (buffer-name) (sys-name)))) (title name) (icon-name (buffer-name)))))
(global-set-key "\C-zb" 'find-buffer-other-screen)
(global-set-key "\C-z\C-f" 'find-file-other-screen)
(global-set-key "\C-z\C-m" (definteractive (unmap-screen (minibuf-screen)) (sleep-for 1) (map-screen (minibuf-screen))))
(global-set-key "\C-zm" (definteractive (change-screen-size 73 1 (minibuf-screen)) (raise-screen (minibuf-screen))))
(global-set-key "\C-zM" (definteractive (change-screen-size 140 1 (minibuf-screen)) (raise-screen (minibuf-screen))))
(global-set-key "\C-z5" 'split-screen-vertically)

;;; For calc:
(defvar calc-mode-mouse-map (create-mouse-map))
(setq calc-mode-hook
      (function (lambda () (use-local-mouse-map calc-mode-mouse-map))))

;;; Function key definitions:
(require 'my-sun3-keys)

(define-function-key "MS-L2" 'remove-buffer-from-screen)
(define-function-key "S-L3" "\C-Z4\C-F")
(define-function-key "CS-L3" "\C-Z4b")

(define-function-key "S-F1" 'describe-mouse)
(define-function-key "CS-F1" 'mouse-helper)
(define-function-key "S-F2" 'describe-mouse-briefly)
(define-function-key "S-F8" 'prev-switch-screen)
(define-function-key "S-F9" 'switch-screen)

(define-function-key "S-R1" 'scroll-other-screen-down-line)
(define-function-key "S-R2" 'scroll-other-screen-down)
(define-function-key "S-R3" 'beginning-of-buffer-other-screen)
(define-function-key "S-R4" 'scroll-other-screen-up-line)
(define-function-key "S-R5" 'scroll-other-screen-up)
(define-function-key "S-R6" 'end-of-buffer-other-screen)
(define-function-key "MS-Up" "\e\e\C-P")
(define-function-key "MS-Down" "\e\e\C-N")

;;; Mouse bindings:

;; LEFT: select-buffer allowing drag and mode-line drag:
(global-set-mouse mouse-left mouse-down 'start-mouse-drag)
(global-set-mouse mouse-left mouse-up 'end-mouse-drag)
(global-set-mouse mouse-mode-left mouse-down nil)
(global-set-mouse mouse-mode-left mouse-up 'mouse-select-buffer)

;; MIDDLE: extend drag; mode, mark-whole-buffer:
(global-set-mouse mouse-middle mouse-down 'extend-mouse-drag)
(global-set-mouse mouse-middle mouse-up 'end-mouse-drag)
(global-set-mouse mouse-mode-middle mouse-down (mousefun-set-point
						(mark-whole-buffer)))

;; RIGHT: play macro:
(global-set-mouse mouse-right mouse-down (mousefun-set-point
					  (undo-boundary)
					  (setq last-command nil)
					  (call-last-kbd-macro)))
(global-set-mouse mouse-right mouse-up nil)
(global-set-mouse mouse-mode-right mouse-down (mousefun-set-point 
					       (undo-boundary)
					       (setq last-command nil)
					       (call-last-kbd-macro)))
(global-set-mouse mouse-minibuf-right mouse-down (mousefun
						  (undo-boundary)
						  (setq last-command nil)
						  (call-last-kbd-macro)))

;; SHIFT-LEFT: vertical scroll down/up when mouse in top/bottom of window:
(global-set-mouse mouse-left mouse-shift nil)
(global-set-mouse mouse-left mouse-shift-up 'mouse-vscroll)

;; CONTROL-LEFT: isearch thing at mouse:
(global-set-mouse mouse-left mouse-control 'mouse-select-thing)
(global-set-mouse mouse-left mouse-control-up 'mouse-isearch-thing)

;; CONTROL-RIGHT: calc-grab thing at point:
(global-set-mouse mouse-right mouse-control 'mouse-select-thing)
(global-set-mouse mouse-right mouse-control-up 'mouse-calc-grab-thing)

;; CONTROL-SHIFT-LEFT: pop-point in window:
(global-set-mouse mouse-left mouse-control-shift (mousefun-set-point (pop-point)))
(global-set-mouse mouse-mode-left mouse-control-shift (mousefun-set-point (pop-point)))

;; META-CONTROL-LEFT: pop-bmark in window:
(global-set-mouse mouse-left mouse-meta-control (mousefun-set-point (pop-bmark)))
(global-set-mouse mouse-mode-left mouse-meta-control (mousefun-set-point (pop-bmark)))

;; META-SHIFT-LEFT: grab thing at mouse:
(global-set-mouse mouse-left mouse-meta-shift 'mouse-select-thing)
(global-set-mouse mouse-left mouse-meta-shift-up 'mouse-grab-thing)

;; META-SHIFT-MIDDLE: kill thing at mouse:
(global-set-mouse mouse-middle mouse-meta-shift 'mouse-select-thing)
(global-set-mouse mouse-middle mouse-meta-shift-up 'mouse-kill-thing)

;; META-SHIFT-RIGHT: copy thing at mouse:
(global-set-mouse mouse-right mouse-meta-shift 'mouse-select-thing)
(global-set-mouse mouse-right mouse-meta-shift-up 'mouse-copy-thing)

;; SHIFT_MIDDLE: copy dragged region:
(global-set-mouse mouse-middle mouse-shift 'start-mouse-grab-drag)
(global-set-mouse mouse-middle mouse-shift-up 'mouse-copy-drag)

;; CONTROL-MIDDLE: wipe dragged region:
(global-set-mouse mouse-middle mouse-control 'start-mouse-grab-drag)
(global-set-mouse mouse-middle mouse-control-up 'mouse-kill-drag)

;; CONTROL-SHIFT-MIDDLE: move region to mouse:
(global-set-mouse mouse-middle mouse-control-shift 'start-mouse-grab-drag)
(global-set-mouse mouse-middle mouse-control-shift-up 'mouse-move-drag)

;; CONTROL-SHIFT-RIGHT: copy region to mouse:
(global-set-mouse mouse-right mouse-control-shift 'start-mouse-grab-drag)
(global-set-mouse mouse-right mouse-control-shift-up 'mouse-grab-drag)

;; SHIFT-RIGHT: yank to mouse:
(global-set-mouse mouse-right mouse-shift 'mouse-set-point)
(global-set-mouse mouse-right mouse-shift-up 'mouse-yank)
(global-set-mouse mouse-mode-right mouse-shift 'mouse-set-point)
(global-set-mouse mouse-mode-right mouse-shift-up 'mouse-yank)

;; SHIFT-MODE-LEFT/MIDDLE/RIGHT: Make the mode line a simulated scroll bar:
(global-set-mouse mouse-mode-left mouse-shift
		  (mousefun-set-point (scroll-up (/ (* (window-height) 
						       (- mouse::x (car (window-edges))))
						    (window-width)))))
(global-set-mouse mouse-mode-middle mouse-shift
		  (mousefun-set-point (goto-char (/ (* (point-max) 
						       (- mouse::x (car (window-edges))))
						    (window-width)))))
(global-set-mouse mouse-mode-right mouse-shift
		  (mousefun-set-point (scroll-down (/ (* (window-height)
							 (- mouse::x (car (window-edges))))
						      (window-width)))))

;; META-LEFT: find-tag at point; mode, find next tag; mini, find-tag interactively:
(global-set-mouse mouse-left mouse-meta (mousefun-set-point (find-tag (find-tag-default))))
(global-set-mouse mouse-mode-left mouse-meta (mousefun-set-point (tags-loop-continue nil)))
(global-set-mouse mouse-minibuf-left mouse-meta (mousefun (call-interactively 'completing-find-tag)))

;; META-CONTROL-SHIFT-LEFT: split window vertically; mode, split vertically in half:
(global-set-mouse mouse-left mouse-meta-control-shift 'mouse-split-window)
(global-set-mouse mouse-mode-left mouse-meta-control-shift 'mouse-split-window)

;; META-CONTROL-SHIFT-MIDDLE: delete window:
(global-set-mouse mouse-middle mouse-meta-control-shift (mousefun (delete-window (nth 2 arg))))
(global-set-mouse mouse-mode-middle mouse-meta-control-shift (mousefun (delete-window (nth 2 arg))))

;; META-CONTROL-SHIFT-RIGHT: rename buffer in window:
(global-set-mouse mouse-right mouse-meta-control-shift
		  (mousefun (save-excursion (set-buffer (nth 1 arg))
					    (rename-current-buffer))))
(global-set-mouse mouse-mode-right mouse-meta-control-shift
		  (mousefun (save-excursion (set-buffer (nth 1 arg))
					    (rename-current-buffer))))

;; MINI-LEFT: extended command:
(global-set-mouse mouse-minibuf-left mouse-down 'execute-extended-command)
;; MINI-META-CONTROL-SHIFT-LEFT: eval-expression:
(global-set-mouse mouse-minibuf-left mouse-meta-control-shift (mousefun (call-interactively 'eval-expression)))

;; MINI-MIDDLE: find file / switch-buffer:
(global-set-mouse mouse-minibuf-middle mouse-down (mousefun (call-interactively 'find-file)))
(global-set-mouse mouse-minibuf-middle mouse-meta (mousefun (call-interactively 'find-file-other-window)))
(global-set-mouse mouse-minibuf-middle mouse-shift (mousefun (call-interactively 'find-file-other-screen)))
(global-set-mouse mouse-minibuf-middle mouse-control (mousefun (call-interactively 'switch-to-buffer)))
(global-set-mouse mouse-minibuf-middle mouse-meta-control (mousefun (call-interactively 'switch-to-buffer-other-window)))
(global-set-mouse mouse-minibuf-middle mouse-control-shift (mousefun (call-interactively 'find-buffer-other-screen)))
(global-set-mouse mouse-minibuf-middle mouse-meta-control-shift (mousefun (call-interactively 'electric-buffer-list)))


;;; Functions:

(defvar split-screen-min-width 80
  "Minimum character width of screens created with \\[split-screen-vertically].")
(defvar split-screen-offset 5
  "Pixels to offset screens created with \\[split-screen-vertically].")

(defun split-screen-vertically (&optional arg)
  "Make a new screen half the width of the current one.  With prefix
arg, use that quadrant.
	1 | 4
	--+--
	2 | 3"
  (interactive "P")
  (let* ((width (max split-screen-min-width (/ (screen-width) 2)))
	 (height (/ (screen-height) (if arg 2 1)))
	 (numarg (if arg (prefix-numeric-value arg) 5))
	 (info (screen-information))
	 (font-info (font))
	 (border (nth 5 info))
	 (x (+ (- (car info) border)
	       (if (<= numarg 2)
		   (- split-screen-offset)
		 (+ split-screen-offset (- (nth 2 info) (* width (nth 1 font-info))) border))))
	 (y (+ (- (nth 1 info) border)
	       (if (or (< numarg 2) (> numarg 3))
		   (- split-screen-offset)
		 (+ split-screen-offset (- (nth 3 info) (* height (nth 2 font-info))) border))))
	 (scr (create-screen (current-buffer)
			     (list (cons 'geometry
					 (format "%dx%d+%d+%d" width height x y))))))
    (move-screen x y scr)
    (cursor-to-screen (select-screen scr))))

(defun remove-buffer-from-screen (&optional buf)
  "Remove all instances of BUF (current buffer if nil) from the current
screen and remove the screen from the buffer's screens list."
  (interactive)
  (or buf
      (setq buf (current-buffer)))
  (delete-windows-on buf)
  (save-excursion
    (set-buffer buf)
    (setq allowed-screens (delq (current-screen) allowed-screens))))

;; So minibuffer is not auto-raised:
;(defun select-screen (&optional scr)
;  (interactive)
;  (epoch::select-screen scr))
