;;; Motion-grab functions for epoch. - Kenneth C. Laprade -  1 Nov 90

(defvar mouse-grab-target-point nil "Saved target point for grab-drag.")
(defvar mouse-grab-target-window nil "Window of saved target for grab-drag.")
(defvar mouse-grab-saved-point nil "Saved point in drag buffer for grab-drag.")

(defun start-mouse-grab-drag (arg)
  "Start a mouse drag, saving point.  Use as down button for
mouse-copy-drag, mouse-move-drag, mouse-grab-drag."
  (when arg
    (or mouse-grab-target-point
	(setq mouse-grab-target-point (make-marker)))
    (setq mouse-grab-target-window (selected-window))
    (set-marker mouse-grab-target-point (point))
    (setq mouse::downp 'start)
    (or mouse-grab-saved-point
	(setq mouse-grab-saved-point (make-marker)))
    (set-marker mouse-grab-saved-point (window-point (nth 2 arg)) (nth 1 arg))
    (mouse::set-point arg)
    (set-mouse-marker)
    (setq mouse::last-point (point))
    (if drag-button
	(progn (delete-button drag-button) (setq drag-button nil))))
  (abort-isearch))

(defun mouse-grab-drag (arg &optional delete noinsert nosave setpoint)
  "Copy drag region to point before drag was started.  Should be used as an
up button with start-mouse-grab-drag on the down button.  First ARG is
mouse event.  With optional DELETE, drag region is deleted.  With NOINSERT
region is not inserted at point.  With NOSAVE, region is not saved in kill
ring.  With SETPOINT, point is left at the end of the drag region rather
than its location before the grab."
  (setq mouse::last-point -1)
  (when mouse::downp
    (setq mouse::downp nil)
    (let* ((buf (current-buffer))
	  (start (and (buttonp drag-button) (button-start drag-button)))
	  (end (and start (button-end drag-button)))
	  (end-mark (set-marker (make-marker) end))
	  (region (and start (buffer-substring start end))))
      (and start (not nosave)
	   (progn (setq last-command nil)
		  (copy-region-as-kill start end)
		  (setq this-command nil)))
      (and start delete
	   (progn (undo-boundary)
		  (delete-region start end)))
      (goto-char (marker-position mouse-grab-saved-point))
      (set-marker mouse-grab-saved-point nil)
      (select-screen (screen-of-window mouse-grab-target-window))
      (select-window mouse-grab-target-window)
      (goto-char (marker-position mouse-grab-target-point))
      (set-marker mouse-grab-target-point nil)
      (and start (not noinsert)
	   (progn
	     (or (and delete (equal buf (current-buffer)))
		 (undo-boundary))
	     (insert region)))
      (and start setpoint
	   (progn (select-screen (nth 3 arg))
		  (select-window (nth 2 arg))
		  (goto-char (marker-position end-mark))))
      (if (marker-buffer end-mark)
	  (save-excursion (set-buffer (marker-buffer end-mark))
			  (if drag-button
			      (delete-button drag-button))
			  (setq drag-button nil)))
      (set-marker end-mark nil)))
  (setq last-command nil)
  (epoch::redisplay-screen))

(defun mouse-move-drag (arg)
  "Move drag region to point before drag was started.  Should be
used as an up button with start-mouse-grab-drag on the down button."
  (mouse-grab-drag arg t nil))

(defun mouse-kill-drag (arg)
  "Delete drag region and save a copy in the kill ring.  Moves point to
mouse location.  Should be used as an up button with start-mouse-grab-drag
on the down button."
  (mouse::set-point arg)
  (mouse-grab-drag arg t t nil t))

(defun mouse-copy-drag (arg)
  "Copy drag region to the kill ring.  Should be used as an up button
with start-mouse-grab-drag on the down button."
  (mouse-grab-drag arg nil t))
;;; --------------------------------------------------------------------------
;;; Some sample bindings:
;;;
;(global-set-mouse mouse-middle mouse-shift 'start-mouse-grab-drag)
;(global-set-mouse mouse-middle mouse-shift-up 'mouse-copy-drag)
;(global-set-mouse mouse-middle mouse-control 'start-mouse-grab-drag)
;(global-set-mouse mouse-middle mouse-control-up 'mouse-kill-drag)
;(global-set-mouse mouse-middle mouse-control-shift 'start-mouse-grab-drag)
;(global-set-mouse mouse-middle mouse-control-shift-up 'mouse-move-drag)
;(global-set-mouse mouse-right mouse-shift 'mouse-set-point)
;(global-set-mouse mouse-right mouse-shift-up 'mouse-yank)
;(global-set-mouse mouse-right mouse-control-shift 'start-mouse-grab-drag)
;(global-set-mouse mouse-right mouse-control-shift-up 'mouse-grab-drag)
