;;;
;;; Copyright (C) 1990  Christopher J. Love
;;; Requires Epoch 3.2 or later, and the use of GWM with appropriate 
;;; WOOL patch to the standard fsm. (see comments below)
;;;
;;; This code is distributed in the hope that it will be useful,
;;; bute WITHOUT ANY WARRANTY. No author or distributor accepts
;;; responsibility to anyone for the consequences of using this code
;;; or for whether it serves any particular purpose or works at all,
;;; unless explicitly stated in a written agreement.
;;;
;;; Everyone is granted permission to copy, modify and redistribute
;;; this code, but only under the conditions described in the
;;; GNU Emacs General Public License, except the original author nor his
;;; agents are bound by the License in their use of this code.
;;; (These special rights for the author in no way restrict the rights of
;;;  others given in the License or this prologue)
;;; A copy of this license is supposed to have been given to you along
;;; with Epoch so you can know your rights and responsibilities. 
;;; It should be in a file named COPYING.  Among other things, the
;;; copyright notice and this notice must be preserved on all copies. 
;;;
;;; gwm.el
;;;
;;;	Provides a GWM major mode for editing WOOL files.
;;; Assumes that lisp.el and lisp-mode.el have been loaded.
;;;
;;; To use gwm-mode interactively with Epoch, apply the GWM-MODE-PATCH to your
;;; personal copy of .gwmrc.gwm .  The same changes are listed in the file
;;; gwm-mode.gwm for manual editing.
;;;
(provide 'gwm-mode)
(require 'property)

;;
;; Make this major mode the default for all *.gwm files 
;;
(setq auto-mode-alist
      (append
       '(("\\.gwm" . gwm-mode))
       auto-mode-alist))

;;
;; Handle GWM output coming back, which is flagged by a change to
;; the property GWM-MODE-DATA.
;; 
(push-property "GWM-MODE-DATA" 'handle-gwm-eval-mode)

(defun handle-gwm-eval-mode (prop new scr)
  (select-screen scr)
  (newline)
  (let ((tmp (epoch::get-property "GWM-MODE-DATA")))
    (and (stringp tmp) (insert tmp)))
)

;;
;; Initial version of gwm-mode-syntax-table is same as lisp mode
;;
(defvar gwm-mode-syntax-table () "")
(and (not gwm-mode-syntax-table)
   (setq gwm-mode-syntax-table (copy-syntax-table lisp-mode-syntax-table)))

;;
;; Create gwm-mode key map
;;
(defvar gwm-mode-map () "")
(and (not gwm-mode-map)
  (progn
  (setq gwm-mode-map (make-sparse-keymap))
  (lisp-mode-commands gwm-mode-map)
  (define-key gwm-mode-map "\n" 'eval-wool-sexp)
  (define-key gwm-mode-map "\r" 'newline-and-indent))) 

(defun gwm-mode ()
  "Major mode for editing GWM WOOL code.
Commands:
Blank lines separate paragraphs.  Semicolons start comments.
\\{gwm-mode-map}
Entry to this mode calls the value of gwm-mode-hook"
  (interactive)
  (kill-all-local-variables)
  (use-local-map gwm-mode-map)
  (setq major-mode 'gwm-mode)
  (setq mode-name "GWM mode")
  (lisp-mode-variables nil)
  (resume-property "GWM-MODE-DATA")
  (set-syntax-table gwm-mode-syntax-table)
  (run-hooks 'gwm-mode-hook))

;;
;; eval-wool-sexp and eval-wool-buffer alter contents of cut-buffer, then
;; set property GWM-MODE-EVAL such that gwm will execute the cut-buffer as
;; WOOL code.  Output (with exception of built-in WOOL primitives), will be
;; inserted into current buffer.  (This is vulnerable to speed of X server,
;; etc.)
;;
(defun eval-wool-sexp ()
  "Have GWM evaluate sexp before point."
  (interactive)
  (epoch::set-property "GWM-MODE-EVAL"  (buffer-substring
			    (let ((stab (syntax-table)))
			      (unwind-protect
			      (save-excursion
				(set-syntax-table gwm-mode-syntax-table)
				(forward-sexp -1)
				(point))
			      (set-syntax-table stab)))
			   (point))))

(defun eval-wool-buffer()
  "Have GWM evaluate current buffer."
  (interactive)
  (epoch::set-property "GWM-MODE-EVAL" (buffer-sub-string
					(point-min)
					(point-max))))

