;;;; Common display styles for epoch use.
;;;  Copyright 1990, Dan L. Pierson (pierson@encore.com)
;;;
;;; This file is for use with Epoch, a modified version of GNU Emacs.
;;; Requires Epoch 3.2 or later.
;;;
;;; This code is distributed in the hope that it will be useful,
;;; bute WITHOUT ANY WARRANTY. No author or distributor accepts
;;; responsibility to anyone for the consequences of using this code
;;; or for whether it serves any particular purpose or works at all,
;;; unless explicitly stated in a written agreement.
;;;
;;; Everyone is granted permission to copy, modify and redistribute
;;; this code, but only under the conditions described in the
;;; GNU Emacs General Public License, except the original author nor his
;;; agents are bound by the License in their use of this code.
;;; (These special rights for the author in no way restrict the rights of
;;;  others given in the License or this prologue)
;;; A copy of this license is supposed to have been given to you along
;;; with Epoch so you can know your rights and responsibilities. 
;;; It should be in a file named COPYING.  Among other things, the
;;; copyright notice and this notice must be preserved on all copies. 

(require 'backquote)
(provide 'common-styles)		; get defstyle for compiler
(require 'common-styles)		;  ...
(require 'bitmaps)			; button stipples

;;; A simple style-definition macro
(defmacro defstyle (name &rest body)
  (` (progn (defvar (, name) nil)
	    (if (, name)
		()
		(let ((style (make-style)))
		  (setq (, name) style)
		  (,@ body))))))

;;; Simple display styles
;;; Note that the style names do not contain '*'.  This is because
;;; they should be thought of as constants once this file is loaded.

(defstyle styles:reverse-video
  (set-style-foreground style (background))
  (set-style-background style (foreground)))

(defstyle styles:underline-video
  (set-style-foreground style (foreground))
  (set-style-background style (background))
  (set-style-underline  style  (foreground)))

;;; Stipple display styles

(let ((bitstring "11011101
		  10111011
		  01110111
		  11101110"))
  (defstyle styles:grey-video
    (set-style-foreground         style (foreground))
    (set-style-background         style (background))
    (set-style-stipple            style (string-to-X-bitmap bitstring))
    (set-style-background-stipple style (string-to-X-bitmap- bitstring)))
  (defstyle styles:reverse-grey-video
    (set-style-foreground         style (background))
    (set-style-background         style (foreground))
    (set-style-stipple            style (string-to-X-bitmap bitstring))
    (set-style-background-stipple style (string-to-X-bitmap- bitstring))))

;;; Font-based display styles
;;; You should probably set the font variables to your taste before
;;; loading this file.  These fonts are for my 100dpi display.
(defvar styles:normal-font "-*-courier-medium-r-*-*-20-*-*-*-*-*-*-*")
(defvar styles:bold-font   "-*-courier-bold-r-*-*-20-*-*-*-*-*-*-*")
(defvar styles:italic-font "-*-courier-bold-o-*-*-20-*-*-*-*-*-*-*")

(defstyle styles:bold-video
  (set-style-foreground style (foreground))
  (set-style-background style (background))
  (set-style-font       style styles:bold-font))
(defstyle styles:normal-video
  (set-style-foreground style (foreground))
  (set-style-background style (background))
  (set-style-font       style styles:normal-font))
(defstyle styles:italic-video
  (set-style-foreground style (foreground))
  (set-style-background style (background))
  (set-style-font       style styles:italic-font))

